@@ -4,8 +4,10 @@ const Gio = imports.gi.Gio;
 const Gtk = imports.gi.Gtk;
 const Pango = imports.gi.Pango;
 const Tp = imports.gi.TelepathyGLib;
+const Tpl = imports.gi.TelepathyLogger;
 
 const Lang = imports.lang;
+const Mainloop = imports.mainloop;
 const Notify = imports.notify;
 const Utils = imports.utils;
 
@@ -13,6 +15,12 @@ const MAX_NICK_CHARS = 8;
 const IGNORE_STATUS_TIME = 5;
 const TP_CURRENT_TIME = GLib.MAXUINT32;
 
+const TIMESTAMP_INTERVAL = 300; // seconds of inactivity after which to
+                                // insert a timestamp
+
+const NUM_INITIAL_LOG_EVENTS = 50; // number of log events to fetch on start
+const NUM_LOG_EVENTS = 10; // number of log events to fetch when requesting more
+
 const INDICATOR_OFFSET = 5; // TODO: should be based on line spacing
 
 // Workaround for GtkTextView growing horizontally over time when
@@ -74,7 +82,7 @@ const ChatView = new Lang.Class({
                             Lang.bind(this, this._onStyleUpdated));
 
         this._room = room;
-        this._lastNick = null;
+        this._state = { lastNick: null, lastTimestamp: 0 };
         this._active = false;
         this._toplevelFocus = false;
         this._joinTime = GLib.DateTime.new_now_utc().to_unix();
@@ -82,6 +90,21 @@ const ChatView = new Lang.Class({
         this._hoveringLink = false;
         this._needsIndicator = true;
         this._pending = {};
+        this._pendingLogs = [];
+
+        let isRoom = room.channel.has_interface(Tp.IFACE_CHANNEL_INTERFACE_GROUP);
+        let account = room.channel.connection.get_account();
+        let target = new Tpl.Entity({ type: isRoom ? Tpl.EntityType.ROOM
+                                                   : Tpl.EntityType.CONTACT,
+                                      identifier: room.channel.identifier });
+        let logManager = Tpl.LogManager.dup_singleton();
+        this._logWalker =
+            logManager.walk_filtered_events(account, target,
+                                            Tpl.EventTypeMask.TEXT, null);
+
+        this._fetchingBacklog = true;
+        this._logWalker.get_events_async(NUM_INITIAL_LOG_EVENTS,
+                                         Lang.bind(this, this._onLogEventsReady));
 
         let adj = this.widget.vadjustment;
         this._scrollBottom = adj.upper - adj.page_size;
@@ -93,9 +116,9 @@ const ChatView = new Lang.Class({
 
         let channelSignals = [
             { name: 'message-received',
-              handler: Lang.bind(this, this._insertMessage) },
+              handler: Lang.bind(this, this._insertTpMessage) },
             { name: 'message-sent',
-              handler: Lang.bind(this, this._insertMessage) },
+              handler: Lang.bind(this, this._insertTpMessage) },
             { name: 'pending-message-removed',
               handler: Lang.bind(this, this._pendingMessageRemoved) }
         ];
@@ -125,7 +148,7 @@ const ChatView = new Lang.Class({
 
         room.channel.dup_pending_messages().forEach(Lang.bind(this,
             function(message) {
-                this._insertMessage(room, message);
+                this._insertTpMessage(room, message);
             }));
         this._checkMessages();
     },
@@ -135,18 +158,23 @@ const ChatView = new Lang.Class({
         let tagTable = buffer.get_tag_table();
         let tags = [
           { name: 'nick',
-            pixels_above_lines: 10,
             left_margin: 0 },
+          { name: 'gap',
+            pixels_above_lines: 10 },
           { name: 'message',
             indent: 0 },
           { name: 'highlight',
             weight: Pango.Weight.BOLD },
           { name: 'status',
             left_margin: 0,
-            justification: Gtk.Justification.RIGHT,
-            indent: 0 },
+            indent: 0,
+            justification: Gtk.Justification.RIGHT },
+          { name: 'timestamp',
+            left_margin: 0,
+            indent: 0,
+            weight: Pango.Weight.BOLD,
+            justification: Gtk.Justification.RIGHT },
           { name: 'action',
-            pixels_above_lines: 10,
             left_margin: 0 },
           { name: 'url',
             underline: Pango.Underline.SINGLE
@@ -178,6 +206,8 @@ const ChatView = new Lang.Class({
             foreground_rgba: selectedColor },
           { name: 'status',
             foreground_rgba: dimColor },
+          { name: 'timestamp',
+            foreground_rgba: dimColor },
           { name: 'action',
             foreground_rgba: dimColor },
           { name: 'url',
@@ -216,12 +246,15 @@ const ChatView = new Lang.Class({
         this.widget.connect('unmap', Lang.bind(this, this._updateActive));
         this.widget.connect('state-flags-changed',
                             Lang.bind(this, this._updateToplevel));
+        this.widget.connect('scroll-event', Lang.bind(this ,this._onScroll));
         this.widget.vadjustment.connect('value-changed',
                                  Lang.bind(this, this._checkMessages));
         this.widget.vadjustment.connect('changed',
                                  Lang.bind(this, this._updateScroll));
         this._view.connect('button-release-event',
                            Lang.bind(this, this._handleLinkClicks));
+        this._view.connect('button-press-event',
+                           Lang.bind(this, this._handleLinkClicks));
         this._view.connect('motion-notify-event',
                            Lang.bind(this, this._handleLinkHovers));
     },
@@ -236,6 +269,49 @@ const ChatView = new Lang.Class({
         this._roomSignals = [];
     },
 
+    _onLogEventsReady: function(lw, res) {
+        let [, events] = lw.get_events_finish(res);
+        this._pendingLogs = events.concat(this._pendingLogs);
+        this._insertPendingLogs();
+        this._fetchingBacklog = false;
+    },
+
+    _insertPendingLogs: function() {
+        if (this._pendingLogs.length == 0)
+            return;
+
+        let index = -1;
+        let nick = this._pendingLogs[0].sender.alias;
+        let type = this._pendingLogs[0].message_type;
+        if (!this._logWalker.is_end()) {
+            for (let i = 0; i < this._pendingLogs.length; i++)
+                if (this._pendingLogs[i].sender.alias != nick ||
+                    this._pendingLogs[i].message_type != type) {
+                    index = i;
+                    break;
+                }
+        } else {
+            index = 0;
+        }
+
+        if (index < 0)
+            return;
+
+        let pending = this._pendingLogs.splice(index);
+        let state = { lastNick: null, lastTimestamp: 0 };
+        let iter = this._view.buffer.get_start_iter();
+        for (let i = 0; i < pending.length; i++) {
+            let message = { nick: pending[i].sender.alias,
+                            text: pending[i].message,
+                            timestamp: pending[i].timestamp,
+                            messageType: pending[i].get_message_type(),
+                            shouldHighlight: false };
+            this._insertMessage(iter, message, state);
+            if (!iter.is_end() || i < pending.length - 1)
+                this._view.buffer.insert(iter, '\n', -1);
+        }
+    },
+
     get _nPending() {
         return Object.keys(this._pending).length;
     },
@@ -285,8 +361,29 @@ const ChatView = new Lang.Class({
         this._scrollBottom = adj.upper - adj.page_size;
     },
 
+    _onScroll: function(w, event) {
+        let [, dir] = event.get_scroll_direction();
+
+        if (this._fetchingBacklog ||
+            this.widget.vadjustment.value != 0 ||
+            dir != Gdk.ScrollDirection.UP ||
+            this._logWalker.is_end())
+            return false;
+
+        this._fetchingBacklog = true;
+        Mainloop.timeout_add(500, Lang.bind(this,
+            function() {
+                this._logWalker.get_events_async(NUM_LOG_EVENTS,
+                                                 Lang.bind(this, this._onLogEventsReady));
+                return false;
+            }));
+        return false;
+    },
+
     _pendingMessageRemoved: function(channel, message) {
         let [id,] = message.get_pending_message_id();
+        if (this._pending[id])
+            this._view.buffer.delete_mark(this._pending[id]);
         delete this._pending[id];
     },
 
@@ -295,6 +392,11 @@ const ChatView = new Lang.Class({
         if (button != Gdk.BUTTON_PRIMARY)
             return false;
 
+        let isPress = event.get_event_type() == Gdk.EventType.BUTTON_PRESS;
+
+        if (isPress)
+            this._clickedUrl = null;
+
         let [, eventX, eventY] = event.get_coords();
         let [x, y] = view.window_to_buffer_coords(Gtk.TextWindowType.WIDGET,
                                                   eventX, eventY);
@@ -304,9 +406,15 @@ const ChatView = new Lang.Class({
         for (let i = 0; i < tags.length; i++) {
             let url = tags[i]._url;
             if (url) {
-                if (url.indexOf(':') == -1)
-                    url = 'http://' + url;
-                Gio.AppInfo.launch_default_for_uri(url, null);
+                if (isPress) {
+                    this._clickedUrl = url;
+                    return true;
+                } else if (this._clickedUrl == url) {
+                    if (url.indexOf(':') == -1)
+                        url = 'http://' + url;
+                    Gio.AppInfo.launch_default_for_uri(url, null);
+                    return true;
+                }
                 break;
             }
         }
@@ -401,84 +509,95 @@ const ChatView = new Lang.Class({
         let time = GLib.DateTime.new_now_utc().to_unix();
         if (time - this._joinTime < IGNORE_STATUS_TIME)
             return;
-        this._lastNick = null;
+        this._state.lastNick = null;
         this._ensureNewLine();
-        this._insertWithTagName(text, 'status');
+        let iter = this._view.buffer.get_end_iter();
+        this._insertWithTagName(iter, text, 'status');
     },
 
-    _insertMessage: function(room, message) {
-        let nick = message.sender.alias;
-        let [text, flags] = message.to_text();
+    _formatTimestamp: function(timestamp) {
+        let date = GLib.DateTime.new_from_unix_local(timestamp);
+        let now = GLib.DateTime.new_now_local();
+
+        let daysAgo = now.difference(date) / GLib.TIME_SPAN_DAY;
+
+        let format;
+        // Show only the hour if date is on today
+        if(daysAgo < 1){
+            format = "%H:%M";
+        }
+        // Show the word "Yesterday" and time if date is on yesterday
+        else if(daysAgo <2){
+            /* Translators: this is the word "Yesterday" followed by a time string. i.e. "Yesterday, 14:30"*/
+            // xgettext:no-c-format
+            format = _("Yesterday, %H:%M");
+        }
+        // Show a week day and time if date is in the last week
+        else if (daysAgo < 7) {
+            /* Translators: this is the week day name followed by a time string. i.e. "Monday, 14:30*/
+            // xgettext:no-c-format
+            format = _("%A, %H:%M");
+
+        } else if (date.get_year() == now.get_year()) {
+            /* Translators: this is the month name and day number followed by a time string. i.e. "May 25, 14:30"*/
+            // xgettext:no-c-format
+            format = _("%B %d, %H:%M");
+        } else {
+            /* Translators: this is the month name, day number, year number followed by a time string. i.e. "May 25 2012, 14:30"*/
+            // xgettext:no-c-format
+            format = _("%B %d %Y, %H:%M");
+        }
+
+        return date.format(format);
+    },
+
+    _insertTpMessage: function(room, tpMessage) {
+        let [text, flags] = tpMessage.to_text();
+
+        let message = { nick: tpMessage.sender.alias,
+                        text: text,
+                        messageType: tpMessage.get_message_type() };
+
+        let timestamp = tpMessage.get_sent_timestamp();
+        if (!timestamp)
+            timestamp = tpMessage.get_received_timestamp();
+        message.timestamp = timestamp;
+
+        message.shouldHighlight = this._room.should_highlight_message(tpMessage);
 
         this._ensureNewLine();
 
-        if (nick.length > this._maxNickChars) {
-            this._maxNickChars = nick.length;
-            this._updateIndent();
+        let iter = this._view.buffer.get_end_iter();
+        this._insertMessage(iter, message, this._state);
+
+        let [id, valid] = tpMessage.get_pending_message_id();
+
+        if (message.shouldHighlight && !this._toplevelFocus) {
+            let summary = '%s %s'.format(this._room.display_name, message.nick);
+            let notification = new Notify.Notification(summary, message.text);
+
+            let account = this._room.channel.connection.get_account();
+            let param = GLib.Variant.new('(ssu)',
+                                         [ account.get_object_path(),
+                                           this._room.channel.identifier,
+                                           TP_CURRENT_TIME ]);
+            notification.addAction('default', 'default');
+            notification.connect('action-invoked', function() {
+                let app = Gio.Application.get_default();
+                let action = app.lookup_action('join-room');
+                action.activate(param);
+            });
+            notification.show();
         }
 
-        let tags = [];
-        if (message.get_message_type() == Tp.ChannelTextMessageType.ACTION) {
-            text = "%s %s".format(nick, text);
-            this._lastNick = null;
-            tags.push(this._lookupTag('action'));
-        } else {
-            if (this._lastNick != nick)
-                this._insertWithTagName(nick + '\t', 'nick');
-            this._lastNick = nick;
-            tags.push(this._lookupTag('message'));
-        }
-
-        let shouldHighlight = this._room.should_highlight_message(message);
-        if (shouldHighlight) {
-            tags.push(this._lookupTag('highlight'));
-
-            if (!this._toplevelFocus) {
-                let summary = '%s %s'.format(this._room.display_name, nick);
-                let notification = new Notify.Notification(summary, text);
-
-                let account = this._room.channel.connection.get_account();
-                let param = GLib.Variant.new('(ssu)',
-                                             [ account.get_object_path(),
-                                               this._room.channel.identifier,
-                                               TP_CURRENT_TIME ]);
-                notification.addAction('default', 'default');
-                notification.connect('action-invoked', function() {
-                    let app = Gio.Application.get_default();
-                    let action = app.lookup_action('join-room');
-                    action.activate(param);
-                });
-                notification.show();
-            }
-        }
-
-        let urls = Utils.findUrls(text);
-        let pos = 0;
-        for (let i = 0; i < urls.length; i++) {
-            let url = urls[i];
-            this._insertWithTags(text.substr(pos, url.pos - pos), tags);
-
-            let tag = new Gtk.TextTag();
-            tag._url = url.url;
-            this._view.get_buffer().tag_table.add(tag);
-
-            this._insertWithTags(url.url,
-                                 tags.concat(this._lookupTag('url'), tag));
-
-            pos = url.pos + url.url.length;
-        }
-        this._insertWithTags(text.substr(pos), tags);
-
-
         let buffer = this._view.get_buffer();
-        let [id, valid] = message.get_pending_message_id();
         if (!valid /* outgoing */ ||
             (this._active && this._toplevelFocus && this._nPending == 0)) {
-            this._room.channel.ack_message_async(message, null);
-        } else if (shouldHighlight || this._needsIndicator) {
+            this._room.channel.ack_message_async(tpMessage, null);
+        } else if (message.shouldHighlight || this._needsIndicator) {
             let iter = buffer.get_end_iter();
 
-            if (shouldHighlight) {
+            if (message.shouldHighlight) {
                 let mark = buffer.create_mark(null, iter, true);
                 this._pending[id] = mark;
             }
@@ -496,6 +615,66 @@ const ChatView = new Lang.Class({
         }
     },
 
+    _insertMessage: function(iter, message, state) {
+        let isAction = message.messageType == Tp.ChannelTextMessageType.ACTION;
+        let needsGap = message.nick != state.lastNick || isAction;
+
+        if (message.timestamp - TIMESTAMP_INTERVAL > state.lastTimestamp) {
+            let tags = [this._lookupTag('timestamp')];
+            if (needsGap)
+                tags.push(this._lookupTag('gap'));
+            needsGap = false;
+            this._insertWithTags(iter,
+                                 this._formatTimestamp(message.timestamp) + '\n',
+                                 tags);
+        }
+        state.lastTimestamp = message.timestamp;
+
+        if (message.nick.length > this._maxNickChars) {
+            this._maxNickChars = message.nick.length;
+            this._updateIndent();
+        }
+
+        let tags = [];
+        if (isAction) {
+            message.text = "%s %s".format(message.nick, message.text);
+            state.lastNick = null;
+            tags.push(this._lookupTag('action'));
+            if (needsGap)
+                tags.push(this._lookupTag('gap'));
+        } else {
+            if (state.lastNick != message.nick) {
+                let tags = [this._lookupTag('nick')];
+                if (needsGap)
+                    tags.push(this._lookupTag('gap'));
+                this._insertWithTags(iter, message.nick + '\t', tags);
+            }
+            state.lastNick = message.nick;
+            tags.push(this._lookupTag('message'));
+        }
+
+        if (message.shouldHighlight)
+            tags.push(this._lookupTag('highlight'));
+
+        let text = message.text;
+        let urls = Utils.findUrls(text);
+        let pos = 0;
+        for (let i = 0; i < urls.length; i++) {
+            let url = urls[i];
+            this._insertWithTags(iter, text.substr(pos, url.pos - pos), tags);
+
+            let tag = new Gtk.TextTag();
+            tag._url = url.url;
+            this._view.get_buffer().tag_table.add(tag);
+
+            this._insertWithTags(iter, url.url,
+                                 tags.concat(this._lookupTag('url'), tag));
+
+            pos = url.pos + url.url.length;
+        }
+        this._insertWithTags(iter, text.substr(pos), tags);
+    },
+
     _ensureNewLine: function() {
         let buffer = this._view.get_buffer();
         let iter = buffer.get_end_iter();
@@ -507,19 +686,19 @@ const ChatView = new Lang.Class({
         return this._view.get_buffer().tag_table.lookup(name);
     },
 
-    _insertWithTagName: function(text, name) {
-        this._insertWithTags(text, [this._lookupTag(name)]);
+    _insertWithTagName: function(iter, text, name) {
+        this._insertWithTags(iter, text, [this._lookupTag(name)]);
     },
 
-    _insertWithTags: function(text, tags) {
+    _insertWithTags: function(iter, text, tags) {
         let buffer = this._view.get_buffer();
-        let iter = buffer.get_end_iter();
         let offset = iter.get_offset();
 
         buffer.insert(iter, text, -1);
 
         let start = buffer.get_iter_at_offset(offset);
 
+        buffer.remove_all_tags(start, iter);
         for (let i = 0; i < tags.length; i++)
             buffer.apply_tag(tags[i], start, iter);
     }
